# -*- coding: UTF-8 -*-
"""This module contains all Exception classes to manage errors.
"""
#import traceback, sys

class Error(Exception):
    def __init__(self, message, oerp_traceback=False):
        super(Error, self).__init__()
        self.oerp_traceback = oerp_traceback
        self.message = message
        #self.tb = sys.exc_info()[2]
    def __str__(self):
        #traceback.print_tb(self.tb)
        return u"{message}".format(message=self.message)


class DMSConnectionError(Error):
    pass


class LoginError(Error):
    pass


class ExecuteQueryError(Error):
    pass


class WorkflowQueryError(Error):
    pass


class ReportError(Error):
    pass


class NotAllowedError(Error):
    def __init__(self, message):
        super(NotAllowedError, self).__init__(message)


class UnknownError(Error):
    def __init__(self, message):
        super(UnknownError, self).__init__(message)




# -*- coding: UTF-8 -*-
"""This module defines the :class:`OERP` and :class:`DMS` classes.

The :class:`OERP` class manage the client-side operations which are related to an OpenERP
server. You can use this to write Python programs that performs a variety of
automated jobs that communicate with an OpenERP server.

"""
"""
The :class:`DMS` class manage the Document Management System of an OpenERP server. This one
inherits from `ftplib.FTP <http://docs.python.org/library/ftplib.html>`_ class and has few methods to facilitate some tasks.

"""
import os
import subprocess

from oerplib.oerp import OERP
from oerplib.dms import DMS
from oerplib import error
#
#MSG_ERROR = u"ERREUR"
#MSG_DONE = u"FAIT"
#
#def open_pdf(file_path, pdf_viewer):
#    """Open a PDF file with the PDF Viewer.
#        
#    Keyword arguments:
#    file_path  -- path to the PDF file
#    pdf_viewer -- path to the PDF Viewer executable
#
#    """
#    try:
#        if os.path.exists(file_path):
#            subprocess.Popen(args=[pdf_viewer, file_path])
#    except:
#        raise error.Error(u"Erreur d'ouverture du fichier PDF. \
#Veuillez vérifier si le chemin vers le lecteur PDF est correct.")




# -*- coding: UTF-8 -*-
"""
"""
import collections

from oerplib import factory, error


class ModelPool(collections.MutableMapping):
    """Manage OSV classes. Each OSV class has an object pool (Factory)
    which manage the corresponding objects.

    """

    def __init__(self, oerp):
        super(ModelPool, self).__init__()
        self.oerp = oerp
        self._factories_by_osv_name = {}
        self._factories_by_osv_class = {}

    def get(self, osv_name, refresh=False):
        """Return a factory which is able to create browsable objects
        corresponding to the OSV name supplied.

        """
        if osv_name not in self._factories_by_osv_name:
            facto = factory.Factory(self.oerp, osv_name)
            self._factories_by_osv_name[osv_name] = facto
            self._factories_by_osv_class[facto.osv['class']] = facto
        elif refresh:
            self._factories_by_osv_name[osv_name] = factory.Factory(self.oerp,
                                                                    osv_name)
        return self._factories_by_osv_name[osv_name]

    def get_by_class(self, osv):
        """Return a factory which is able to create browsable objects
        corresponding to the OSV class supplied.

        """
        return self._factories_by_osv_class[osv]

    def __str__(self):
        """Return string representation of this pool."""
        res = {}
        for osv_name, facto in self._factories_by_osv_name.iteritems():
            res[osv_name] = facto.keys()
        return str(res)

    # ---------------------------- #
    # -- MutableMapping methods -- #
    # ---------------------------- #

    def __delitem__(self, osv_name):
        self._factories_by_osv_name[osv_name].clear()
        del self._factories_by_osv_name[osv_name]
        #raise error.NotAllowedError(u"Operation not supported")

    def __getitem__(self, osv_name):
        return self.get(osv_name)

    def __iter__(self):
        for osv_name in self._factories_by_osv_name:
            yield osv_name

    def __len__(self):
        return len(self._factories_by_osv_name)

    def __setitem__(self, osv_name, value):
        raise error.NotAllowedError(u"Operation not supported")




# -*- coding: UTF-8 -*-
from ftplib import FTP
import os
import tempfile

from oerplib.oerp import error

class DMS(FTP, object):
    """ EXPERIMENTAL, NOT TESTED OR DOCUMENTED!
    Represents the Document Management System of an OpenERP server.
    Inherits from FTP class.
    """
    def __init__(self, server, port):
        super(DMS, self).__init__()
        self.server = server
        self.port = port

    def login(self, user="anonymous", pwd="", acct=""):
        try:
            super(DMS, self).connect(self.server, self.port)
        except Exception as exc:
            raise error.DMSConnectionError(u"FTP connection failed. " + \
                                u"Check the DMS service of OpenERP server " + \
                                u"or connection settings")
        try:
            super(DMS, self).login(user, pwd, acct)
        except Exception as exc:
            raise error.DMSConnectionError(u"Wrong login ID or password")

    def get(self, file_name):
        """ Download a file from the OpenERP DMS Service (FTP Server)
            and return its temporary client-side path.

        """
        try:
            (ft_no, ft_name) = tempfile.mkstemp(prefix='oerplib_')
            ft_obj = file(ft_name, 'wb+')
            super(DMS, self).retrbinary('RETR {0}'.format(file_name),
                                        ft_obj.write)
        except:
            raise error.DMSConnectionError(u"The download of the file" + \
                                u"'{0}' has failed.".format(file_name))
        else:
            return ft_name
        finally:
            #outfile.close()
            ft_obj.close()
            os.close(ft_no)
        return None

#    def reset(self, path):
#        """ Reset FTP root directory """
#        #FIXME e.g. path = 'Documents/Production/
#        self.cwd('/{0}/{1}'.format(oerp.config.DATABASE, path))




# -*- coding: UTF-8 -*-
"""This module contains the OERP class which manage the interaction with
the OpenERP server.
"""
import xmlrpclib, socket
import os
import time
import collections
import traceback, sys
import base64, zlib, tempfile

from oerplib import error
from oerplib import factory
from oerplib import pool

class OERP(collections.MutableMapping):
    """Return a new instance of the :class:`OERP` class.
    The optional *database* parameter specifies the default database to use
    when the ``login(user, passwd)`` is called. By default, the port 8069
    is used.

    """
    
#    Represent an OpenERP server.
#    Manage the server connection, requests, objects generation,
#    download of reports and some other client-side functionalities.
#
#    Attributes:
#    user        -- object which contains User's informations
#    server      -- domain name or IP of the server
#    port        -- port used to connect to the server
#    database    -- database name
#
#    Methods:
#    login       -- login an user to OpenERP Server
#    execute     -- executes a XMLRPC request on a model
#    exec_workflow   -- executes a XMLRPC workflow request on a model
#    exec_report -- retrieve a report according to the model supplied
#    browse      -- browse an 'osv' object
#    search      -- search ids of objects that match the supplied criteria
#    read        -- return a data dictionary of the supplied model
#
#    Special notes:
#    May be use as a dictionary to retrieve objects
#    classified by model, then by id, e.g.:
#        order = oerp_instance['sale.order'][42]


    def __init__(self, server, database=None, port=8069):
        self.server = server
        self.port = port
        self.database = self.database_default = database
        self.pool = pool.ModelPool(self)
        self.user = False
        self.xmlrpc_url = 'http://{server}:{port}/xmlrpc'.format(
                server=self.server,
                port=self.port)
        self.sock = xmlrpclib.ServerProxy(self.xmlrpc_url+'/object')
        self.sock_report = xmlrpclib.ServerProxy(self.xmlrpc_url+'/report')
        self.sock_common = xmlrpclib.ServerProxy(self.xmlrpc_url+'/common')

    def login(self, user, passwd, database=None):
        """Log in as the given ``user`` with the password ``passwd`` on the
        database ``database`` and return the corresponding User browsable
        object.
        If ``database`` is not specified, the default one will be used instead.
        If no database is found, a LoginError exception will be raised.

        """
        self.database = database or self.database_default
        if not self.database:
            raise error.LoginError(u"No database specified")
        try:
            user_id = self.sock_common.login(self.database, user, passwd)
        except socket.error as exc:
            raise error.LoginError(u"{0}".format(exc.strerror))
        else:
            if user_id:
                # Create a fake User object just to execute the
                # first query : fetch the real User object
                self.user = type('User', (object,), {
                                    'id': user_id,
                                    'login': user,
                                    'password': passwd,
                                })
                self.user = self.browse('res.users', user_id)
                return self.user
            else:
                raise error.LoginError(u"Wrong login ID or password")

    # ------------------------- #
    # -- Raw XML-RPC methods -- #
    # ------------------------- #

    def execute(self, osv_name, method, *args):
        """Execute a simple XMLRPC method ``method`` on the OSV server class
        ``osv_name``. ``*args`` parameters varies according to the method used.

        """
        if not self.user:
            raise error.LoginError(
                u"Have to be logged to be able to execute queries")
        try:
            return self.sock.execute(
                        self.database, self.user.id,
                        self.user.password, osv_name, method, *args)
        except xmlrpclib.Error as exc:
            raise error.ExecuteQueryError(u"{0}".format(exc.faultCode
                                                        or "Unknown error"),
                                          exc.faultString)

    def exec_workflow(self, *args):
        """XMLRPC Workflow query.

        `WARNING: not sufficiently tested.`

        """
        if not self.user:
            raise error.LoginError(
                u"Have to be logged to be able to execute queries")
        #TODO need to be tested + fix exception
        try:
            return self.sock.exec_workflow(
                        self.database, self.user.id,
                        self.user.password, *args)
        except Exception as exc:
            raise error.WorkflowQueryError(
                        u"Workflow query has failed")

    def exec_report(self, report_name, osv_name, obj_id, report_type='pdf'):
        """Download a report from the OpenERP server via XMLRPC
        and return the path of the file.

        `WARNING: not sufficiently tested.`

        """
        if not self.user:
            raise error.LoginError(
                u"Have to be logged to be able to execute queries")
        data = {'model': osv_name, 'id': obj_id, 'report_type': report_type}
        try:
            report_id = self.sock_report.report(self.database, self.user.id,
                    self.user.password, report_name, [obj_id], data)
        except xmlrpclib.Error as exc:
            raise error.ReportError(u"{0}".format(exc.faultCode))
        state = False
        attempt = 0
        while not state:
            try:
                pdf_data = self.sock_report.report_get(self.database,
                        self.user.id, self.user.password, report_id)
            except xmlrpclib.Error as exc:
                raise error.ReportError(u"{0}".format(exc.faultString))
            state = pdf_data['state']
            if not state:
                time.sleep(1)
                attempt += 1
            if attempt > 200:
                raise error.ReportError(u"Download time exceeded, " + \
                                        u"the operation has been canceled.")
        return self.__print_file_data(pdf_data)

    @staticmethod
    def __print_file_data(data):
        """Print data in a temporary file and return the path of this one."""
        if 'result' not in data:
            raise error.ReportError(
                    u"Invalid data, the operation has been canceled.")
        content = base64.decodestring(data['result'])
        if data.get('code') == 'zlib':
            content = zlib.decompress(content)
    
        if data['format'] in ['pdf', 'html', 'doc', 'xls',
                              'sxw', 'odt', 'tiff']:
            if data['format'] == 'html' and os.name == 'nt':
                data['format'] = 'doc'
            (file_no, file_path) = tempfile.mkstemp('.'+data['format'],
                                                    'oerplib_')
            with file(file_path, 'wb+') as fp:
                fp.write(content)
            os.close(file_no)
            return file_path

    # ------------------------- #
    # -- High Level methods  -- #
    # ------------------------- #

    def browse(self, osv_name, ids, join=False, refresh=False):
        """Return a browsable object (or a list of objects)
        according to the OSV name and ID (or IDs) supplied.
        The ``join`` parameter allow to get at the same time the first level
        of relations (Many2One, One2Many and Many2Many).
        ``refresh`` option will reinitialize the object if this one has already
        been loaded previously.

        """
        if isinstance(ids, list):
            return [self.browse(osv_name, o_id, join, refresh)
                    for o_id in ids]
        else:
            return self.pool.get(osv_name).generate_browse(ids, join, refresh)

    def search(self, osv_name, args):
        """Return a list of IDs of records matching the given criteria in
        ``args`` parameter. ``args`` must be of the form
        ``[('name', '=', 'John'), (...)]``

        """
        return self.execute(osv_name, 'search', args)

    def create(self, osv_name, vals):
        """Create a new record with the specified values contained in the
        ``vals`` dictionary (e.g. ``{'name': 'John', ...}``).
        Return the ID of the new record.

        """
        return self.execute(osv_name, 'create', vals)

    def read(self, osv_name, ids, fields=[]):
        """Return the ID of each record with the values
        of the requested fields ``fields`` from the OSV server class
        ``osv_name``. If ``fields`` is not specified, all fields values
        will be retrieved.

        """
        return self.execute(osv_name, 'read', ids, fields)

    def write(self, osv, ids=[], vals={}):
        """Update records with given IDs (e.g. ``[1, 42, ...]``)
        with the given values contained in the ``vals`` dictionary
        (e.g. ``{'name': 'John', ...}``).
        ``osv`` parameter may be the OSV server class name
        (e.g. ``'sale.order'``) or an ``OSV`` instance (browsable object).
        In the latter case, ``vals`` will automatically contain all the changed
        values of the object and ``ids`` the object ID, thus the parameters
        ``ids`` and ``vals`` are useless.
        Return True.

        """
        if isinstance(osv, factory.OSV):
            return self.pool.get_by_class(osv.__class__).write(osv)
        return self.execute(osv, 'write', ids, vals)

    def unlink(self, osv, ids):
        """Delete records with the given IDs (e.g. ``[1, 42, ...]``).
        ``osv`` parameter may be the OSV server class name
        (e.g. ``'sale.order'``) or an OSV instance (browsable object).
        In the latter case, the parameter ``ids`` is useless.
        Return True.

        """
        if isinstance(osv, factory.OSV):
            return self.pool.get_by_class(osv.__class__).unlink(osv)
        return self.execute(osv, 'unlink', ids)

    def refresh(self, osv_obj):
        """Restore original values of the object ``osv_obj`` from data
        retrieved on the OpenERP server.
        Thus, all changes made locally on the object are canceled.

        """
        return self.pool.get_by_class(osv_obj.__class__).refresh(osv_obj)

    def reset(self, osv_obj):
        """Cancel all changes made locally on the object ``osv_obj``.
        No request to the server is executed to perform this operation.
        Therefore, values restored may be outdated.

        """
        return self.pool.get_by_class(osv_obj.__class__).reset(osv_obj)

    def get_osv_name(self, osv_obj):
        """Return the OSV name of the OSV instance ``osv_obj`` supplied."""
        if not isinstance(osv_obj, factory.OSV):
            raise ValueError(u"Value is not an instance of OSV class")
        return self.pool.get_by_class(osv_obj.__class__).osv['name']

    # ---------------------------- #
    # -- MutableMapping methods -- #
    # ---------------------------- #

    def __delitem__(self, osv_name):
        del self.pool[osv_name]
        #raise error.NotAllowedError(u"Operation not supported")

    def __getitem__(self, osv_name):
        return self.pool[osv_name]

    def __iter__(self):
        for osv_name in self.pool:
            yield osv_name

    def __len__(self):
        return len(self.pool)

    def __setitem__(self, osv_name, value):
        raise error.NotAllowedError(u"Operation not supported")




#!/usr/bin/env python
import oerplib

# XMLRPC server configuration (NETRPC is not supported)
XMLRPC_SERVER = 'localhost'
XMLRPC_SERVER_PORT = 8069
# DMS service configuration
FTP_SERVER = 'localhost'
FTP_SERVER_PORT = 8021
# Name of the OpenERP database to use
DATABASE = 'database-dev'
# Use OpenERP Document Management System module
# (this one must be installed on the server side)
USE_DMS = True

USER = 'admin'
PASSWORD = 'password'

try:
    # Login
    oerp = oerplib.OERP(
            server=XMLRPC_SERVER,
            port=XMLRPC_SERVER_PORT,
            database=DATABASE,
            )
    oerp.login(USER, PASSWORD)

    # ----------------------- #
    # -- Low level methods -- #
    # ----------------------- #

    # Execute - search
    user_ids = oerp.execute('res.users', 'search', [('id', '=', oerp.user.id)])

    # Execute - read
    user_data = oerp.execute('res.users', 'read', user_ids[0])

    # Execute - write
    oerp.execute('res.users', 'write', user_ids[0], {'name': u"Administrator"})

    # Execute - create
    #FIXME

    # Get a report
    pdf_path = oerp.exec_report('sale.order', 'sale.order', 1)

    # --------------------------- #
    # -- Search IDs of objects -- #
    # --------------------------- #

    # Search IDs of a model that match criteria
    assert oerp.user.id in oerp.search('res.users', [('name', 'ilike', u"Administrator"),])

    # ---------------------------- #
    # -- Read data of an object -- #
    # ---------------------------- #
    print(oerp.read('res.users', oerp.user.id))

    # -------------------- #
    # -- Browse objects -- #
    # -------------------- #

    # Browse an object 
    user = oerp.browse('res.users', oerp.user.id, join=True)
    print(user.name)
    print(user.company_id.name)
    # Note: 'join' parameter allow to fetch related objects at the same time
    # which avoids another XML-RPC request when these ones are accessed

    # .. or many objects
    for order in oerp.browse('sale.order', [68, 69]):
        print(order.name)
        print(order.partner_id.name)
        for line in order.order_line:
            print('\t{0}'.format(line.name))

    # ------------------------------- #
    # -- OERP Dictionary interface -- #
    # ------------------------------- #

    # An object may be accessed via a dictionary interface
    print(oerp['res.users'][oerp.user.id].name)

    # Modifications can be saved too
    oerp['res.users'][oerp.user.id].name = u"James Bond"
    oerp.write(oerp['res.users'][oerp.user.id])

    # Changes can be canceled
    oerp['res.users'][oerp.user.id].name = u"Thierry la Fronde"
    oerp.reset(oerp['res.users'][oerp.user.id])
    print(oerp['res.users'][oerp.user.id].name) # Display "James Bond"

    # And they can be updated from Open ERP server
    oerp.refresh(oerp['res.users'][oerp.user.id])

    # Many2One fields can be updated easily (with an ID or with an object)
    oerp['res.users'][oerp.user.id].address_id = oerp.browse('res.partner.address', 1)
    oerp.write(oerp['res.users'][oerp.user.id])
    oerp['res.users'][oerp.user.id].address_id = 1  # ID supplied: same effect as above, implicit relation used
    oerp.write(oerp['res.users'][oerp.user.id])

    # -------------------------------- #
    # -- Document Management System -- #
    # -------------------------------- #

    if USE_DMS:
        # Get a DMS object
        dms = oerplib.DMS(FTP_SERVER, FTP_SERVER_PORT)
        # Login on the DMS
        dms.login(USER, PASSWORD)
        print("---- Content of {0}".format(dms.pwd()))
        dms.dir()
        dms.cwd(DATABASE)
        print("---- Content of {0}".format(dms.pwd()))
        dms.dir()
        dms.cwd('Documents')
        print("---- Content of {0}".format(dms.pwd()))
        dms.dir()
        dms.cwd('Sales Order')
        print("---- Content of {0}".format(dms.pwd()))
        dms.dir()

        dms.cwd('All Sales Order/SO001')
        pdf_path = dms.get('SO001_print.pdf')

except oerplib.error.Error as e:
    print(e)
except Exception as e:
    print(e)




# -*- coding: UTF-8 -*-
"""This module contains classes representing the fields supported by OpenObject.
"""
import datetime

from oerplib import error
OSV_CLASS = None


class BaseField(object):
    """Field which all other fields inherit.
    Manage common metadata.
    """
    def __init__(self, factory, name, data):
        self.factory = factory
        self.name = name
        self.type = 'type' in data and data['type'] or False
        self.string = 'string' in data and data['string'] or False
        self.size = 'size' in data and data['size'] or False
        self.required = 'required' in data and data['required'] or False
        self.readonly = 'readonly' in data and data['readonly'] or False
        self.help = 'help' in data and data['help'] or False
        self.states = 'states' in data and data['states'] or False

    def get_property(self):
        """Generate a Python class property corresponding to the field type."""
        pass

    def check_value(self, value):
        """Check the validity of a value for the field."""
        if self.readonly:
            raise error.NotAllowedError(
                u"'{field_name}' field is readonly".format(
                    field_name=self.name))
        if self.size and len(value) > self.size:
            raise error.NotAllowedError(
                u"Lenght of the '{field_name}' is limited to {size}".format(
                    field_name=self.name,
                    size=self.size))
        if not value and self.required:
            raise error.NotAllowedError(
                u"'{field_name}' field require a value".format(
                    field_name=self.name))
        return value


class SelectionField(BaseField):
    """Represent the OpenObject 'fields.selection'"""
    def __init__(self, factory, name, data):
        super(SelectionField, self).__init__(factory, name, data)
        self.selection = 'selection' in data and data['selection'] or False

    def get_property(self):
        def getter(obj):
            return getattr(obj, "_{0}".format(self.name))

        def setter(obj, value):
            value = self.check_value(value)
            setattr(obj, "_{0}".format(self.name), value)
            self.factory.objects[obj.id]['fields_updated'].append(self.name)

        return property(getter, setter)

    def check_value(self, value):
        super(SelectionField, self).check_value(value)
        selection = [val[0] for val in self.selection]
        if value not in selection:
            raise error.NotAllowedError(
                u"The value '{value}' supplied doesn't match with the possible \
values '{selection}' for the '{field_name}' field".format(
                    value=value,
                    selection=selection,
                    field_name=self.name,
            ))
        return value


class Many2ManyField(BaseField):
    """Represent the OpenObject 'fields.many2many'"""
    def __init__(self, factory, name, data):
        super(Many2ManyField, self).__init__(factory, name, data)
        self.relation = 'relation' in data and data['relation'] or False
        self.context = 'context' in data and data['context'] or False
        self.domain = 'domain' in data and data['domain'] or False

    def get_property(self):
        def getter(obj):
            return [self.factory.oerp.browse(self.relation, o_id) 
                    for o_id in self.factory.objects[obj.id]['raw_data'][self.name]]

        def setter(obj, value):
            raise error.NotAllowedError(u"Not implemented yet")

        return property(getter, setter)

    def check_value(self, value):
        #TODO Many2Many.check_value()
        super(Many2ManyField, self).check_value(value)

class Many2OneField(BaseField):
    """Represent the OpenObject 'fields.many2one'"""
    def __init__(self, factory, name, data):
        super(Many2OneField, self).__init__(factory, name, data)
        self.relation = 'relation' in data and data['relation'] or False
        self.context = 'context' in data and data['context'] or False
        self.domain = 'domain' in data and data['domain'] or False

    def get_property(self):
        def is_int(value):
            try:
                int(value)
                return True
            except Exception as exc:
                return False

        def getter(obj):
            if getattr(obj, "_{0}".format(self.name)):
                return obj.__class__._oerp.browse(
                    self.relation,
                    getattr(obj, "_{0}".format(self.name))[0]
                )

        def setter(obj, value):
            if isinstance(value, OSV_CLASS):
                o_rel = value
            if is_int(value):
                o_rel = obj.__class__._oerp.browse(self.relation, value)
            o_rel = self.check_value(o_rel)
            setattr(obj, "_{0}".format(self.name), [o_rel.id, o_rel.name])
            self.factory.objects[obj.id]['fields_updated'].append(self.name)

        return property(getter, setter)

    def check_value(self, value):
        super(Many2OneField, self).check_value(value)
        oerp = self.factory.oerp
        value_factory = oerp.pool.get_by_class(value.__class__)
        if value_factory.osv['name'] != self.relation:
            raise error.NotAllowedError(
                (u"Instance of '{osv_name}' supplied doesn't match with the " +\
                u"relation '{relation}' of the '{field_name}' field.").format(
                    osv_name=value_factory.osv['name'],
                    relation=self.relation,
                    field_name=self.name))
        return value

class One2ManyField(BaseField):
    """Represent the OpenObject 'fields.one2many'"""
    def __init__(self, factory, name, data):
        super(One2ManyField, self).__init__(factory, name, data)
        self.relation = 'relation' in data and data['relation'] or False
        self.context = 'context' in data and data['context'] or False
        self.domain = 'domain' in data and data['domain'] or False

    def get_property(self):
        def getter(obj):
            return [self.factory.oerp.browse(self.relation, o_id) 
                    for o_id in self.factory.objects[obj.id]['raw_data'][self.name]]

        def setter(obj, value):
            #TODO: setter of One2Many field
            raise error.NotAllowedError(u"Not implemented yet")

        return property(getter, setter)

    def check_value(self, value):
        #TODO One2Many.check_value()
        super(One2ManyField, self).check_value(value)


class DateField(BaseField):
    """Represent the OpenObject 'fields.data'"""
    pattern = "%Y-%m-%d"
    def __init__(self, factory, name, data):
        super(DateField, self).__init__(factory, name, data)

    def get_property(self):
        def getter(obj):
            value = getattr(obj, "_{0}".format(self.name))
            try:
                res = datetime.datetime.strptime(value, self.pattern).date()
            except Exception:
                res = value
            return res

        def setter(obj, value):
            value = self.check_value(value)
            setattr(obj, "_{0}".format(self.name), value)
            self.factory.objects[obj.id]['fields_updated'].append(self.name)

        return property(getter, setter)

    def check_value(self, value):
        super(DateField, self).check_value(value)
        if isinstance(value, datetime.date):
            value = value.strftime("%Y-%m-%d")
        elif isinstance(value, basestring):
            try:
                datetime.datetime.strptime(value, self.pattern)
            except:
                raise ValueError(
                    "Value not well formatted, expecting '{0}' format".format(
                        self.pattern))
        else:
            raise ValueError("Expecting a datetime.date object or basestring")
        return value


class DateTimeField(BaseField):
    """Represent the OpenObject 'fields.datetime'"""
    pattern = "%Y-%m-%d %H:%M:%S"
    def __init__(self, factory, name, data):
        super(DateTimeField, self).__init__(factory, name, data)

    def get_property(self):
        def getter(obj):
            value = getattr(obj, "_{0}".format(self.name))
            try:
                res = datetime.datetime.strptime(value, self.pattern)
            except Exception:
                res = value
            return res

        def setter(obj, value):
            value = self.check_value(value)
            setattr(obj, "_{0}".format(self.name), value)
            self.factory.objects[obj.id]['fields_updated'].append(self.name)

        return property(getter, setter)

    def check_value(self, value):
        super(DateTimeField, self).check_value(value)
        if isinstance(value, datetime.datetime):
            value = value.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(value, basestring):
            try:
                datetime.datetime.strptime(value, self.pattern)
            except:
                raise ValueError(
                    "Value not well formatted, expecting '{0}' format".format(
                        self.pattern))
        else:
            raise ValueError(
                "Expecting a datetime.datetime object or basestring")
        return value


class ValueField(BaseField):
    """Represent simple OpenObject fields:
    - 'fields.char',
    - 'fields.float',
    - 'fields.integer',
    - 'fields.boolean',
    - 'fields.text',
    - 'fields.binary',
    """
    def __init__(self, factory, name, data):
        super(ValueField, self).__init__(factory, name, data)

    def get_property(self):
        def getter(obj):
            return getattr(obj, "_{0}".format(self.name))

        def setter(obj, value):
            value = self.check_value(value)
            setattr(obj, "_{0}".format(self.name), value)
            self.factory.objects[obj.id]['fields_updated'].append(self.name)

        return property(getter, setter)


def generate_field(factory, name, data):
    """Generate a well-typed field according to the data dictionary supplied
    (obtained via 'fields_get' XML-RPC/NET-RPC method).

    """
    assert 'type' in data
    field = None
    if data['type'] == 'selection':
        field = SelectionField(factory, name, data)
    elif data['type'] == 'many2many':
        field = Many2ManyField(factory, name, data)
    elif data['type'] == 'many2one':
        field = Many2OneField(factory, name, data)
    elif data['type'] == 'one2many':
        field = One2ManyField(factory, name, data)
    elif data['type'] == 'date':
        field = DateField(factory, name, data)
    elif data['type'] == 'datetime':
        field = DateTimeField(factory, name, data)
    elif data['type'] in ['char', 'float', 'integer',
                          'boolean', 'text', 'binary']:
        field = ValueField(factory, name, data)
    else:
        raise error.UnknownError(
            u"Can't instanciate the field '{field_name}', type unknown".format(
                field_name=name))
    return field




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from distutils.core import setup

name = 'OERPLib'
version = '0.1'
release = '0.1.2'
description = 'Python library which allows to easily interact with an OpenERP server.'
author = u"ABF Osiell - Sébastien Alix"
author_email = 'sebastien.alix@osiell.com'
url = 'http://www.osiell.com/'
#download_url = 'http://pypi.python.org/packages/source/o/oerplib/oerplib-%s.tar.gz' % version,
download_url = 'https://bitbucket.org/SebAlix/python-oerplib/downloads'
license = 'LGPL v3'

cmdclass = {}
command_options = {}
# 'build_sphinx' option
try:
    from sphinx.setup_command import BuildDoc
    cmdclass = {'build_sphinx': BuildDoc}
    command_options={
        'build_sphinx': {
            #'project': ('setup.py', name),
            'version': ('setup.py', version),
            'release': ('setup.py', release),
            'source_dir': ('setup.py', 'doc/source'),
            'build_dir': ('setup.py', 'doc/'),
         }}
except Exception:
    print("No Sphinx module found. You have to install Sphinx to be able to generate the documentation.")

setup(name=name,
      version=release,
      description=description,
      long_description=open('README.txt').read(),
      author=author,
      author_email=author_email,
      url=url,
      download_url=download_url,
      packages=['oerplib', 'oerplib.test'],
      license=license,
      cmdclass=cmdclass,
      command_options=command_options,
      )




# -*- coding: UTF-8 -*-
"""This module provides the OSV class
and some functions to manipulate its instances.
"""
import collections
from oerplib import error, fields


class OSV(object):
    """Base class that all OSV classes inherit from."""
    def __init__(self, factory, o_id):
        self.id = o_id
        factory.refresh(self)

fields.OSV_CLASS = OSV


class Factory(collections.MutableMapping):
    """Manage the objects corresponding to a OSV class."""

    fields_reserved = ['id', 'oerp']
    def __init__(self, oerp, osv_name):
        super(Factory, self).__init__()
        self.oerp = oerp
        self.osv = {
            'name': osv_name,
            'class': None,
            'fields': {},
        }
        self.osv['class'] = self.generate_osv(osv_name)
        self.objects = {}

    def generate_browse(self, obj_id, join=False, refresh=False):
        """Generate an instance of the OSV class."""
        # TODO be able to refresh all joined relations
        if obj_id not in self.objects or refresh:
            self.objects[obj_id] = {
                'raw_data': None,
                'fields_updated': [],
                'instance': None,
                'join': join,
            }
            try:
                self.objects[obj_id]['instance'] = self.osv['class'](self, obj_id)
            except error.ExecuteQueryError as exc:
                del self.objects[obj_id]
                raise error.ExecuteQueryError(
                    u"There is no '{0}' record with ID {1}.".format(
                        self.osv['name'], obj_id))

        return self.objects[obj_id]['instance']

    def generate_osv(self, osv_name):
        """Generate a class (with all its fields)
        corresponding to the OSV name supplied.

        """
        try:
            fields_get = self.oerp.execute(osv_name, 'fields_get')
        except error.ExecuteQueryError as exc:
            raise error.ExecuteQueryError(
                u"There is no OSV class named '{0}'.".format(self.osv['name']))
        cls_name = osv_name.replace('.', '_')
        for field_name, field_data in fields_get.items():
            if field_name not in self.fields_reserved:
                self.osv['fields'][field_name] = fields.generate_field(
                                                            self,
                                                            field_name,
                                                            field_data)
        cls = type(cls_name, (OSV,), {})
        cls._oerp = self.oerp
        return cls

    def write(self, obj):
        """Send values of fields updated to the OpenERP server."""
        obj_info = self.objects[obj.id]
        vals = {}
        for field_name in obj_info['fields_updated']:
            if field_name in obj_info['raw_data']:
                field = self.osv['fields'][field_name]
                # Many2One fields
                if isinstance(field, fields.Many2OneField):
                    vals[field_name] = getattr(obj,
                                               "_{0}".format(field_name))[0]
                # One2Many fields
                elif isinstance(field, fields.One2ManyField):
                    pass #TODO write One2Many value
                # Many2Many fields
                elif isinstance(field, fields.Many2ManyField):
                    pass #TODO write Many2Many value
                # All other fields
                else:
                    vals[field_name] = getattr(obj, "_{0}".format(field_name))
        try:
            res = self.oerp.write(self.osv['name'], [obj.id], vals)
        except error.Error as exc:
            raise
        else:
            # Update raw_data dictionary
            self.refresh(obj)
            return res

    def refresh(self, obj):
        """Retrieve field values from OpenERP server.
        May be used to restore the original values
        in the purpose to cancel all changes made.

        """
        obj_info = self.objects[obj.id]
        obj_info['raw_data'] = self.oerp.read(self.osv['name'], obj.id)
        self.reset(obj)

    def reset(self, obj):
        """Cancel all changes by restoring field values with original values
        obtained during the last refresh (object instanciation or
        last call to refresh() method).

        """
        obj_info = self.objects[obj.id]
        obj_info['fields_updated'] = []
        # Load fields
        for field in self.osv['fields'].values():
            if field.name in obj_info['raw_data']:
                setattr(obj, "_{0}".format(field.name),
                        obj_info['raw_data'][field.name])
                setattr(obj.__class__, field.name,
                        field.get_property())
                # Preload value (for relation fields as One2Many, Many2One...)
                if obj_info['join']:
                    getattr(obj, field.name)

    def unlink(self, obj):
        """Delete the object locally and from the server."""
        #TODO delete locally
        return self.oerp.unlink(self.osv['name'], [obj.id])

    def __str__(self):
        """Return string representation of this factory."""
        return str([obj_id for obj_id in self.objects.keys()])

    # ---------------------------- #
    # -- MutableMapping methods -- #
    # ---------------------------- #

    def __delitem__(self, obj_id):
        del self.objects[obj_id]
        #raise error.NotAllowedError(u"Operation not supported")

    def __getitem__(self, obj_id):
        return self.generate_browse(obj_id)

    def __iter__(self):
        for obj_id in self.objects:
            yield obj_id

    def __len__(self):
        return len(self.objects)

    def __setitem__(self, key, value):
        raise error.NotAllowedError(u"Operation not supported")

